// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

#![cfg(todo)]

use hashbrown::{hash_map::Entry, HashMap};
use num_bigint::BigInt;
use petgraph::{
    graph::{NodeIndex, UnGraph},
    unionfind::UnionFind,
};

use crate::{
    annotations::{Annotation, TargetedAnnotation},
    bundle::{BundleField, BundleType},
    enum_::{EnumType, EnumVariant},
    expr::{
        ops,
        target::{
            Target, TargetBase, TargetChild, TargetPathArrayElement, TargetPathBundleField,
            TargetPathDynArrayElement, TargetPathElement,
        },
        ExprEnum, Flow,
    },
    intern::{Intern, Interned, Memoize},
    memory::{DynPortType, MemPort},
    module::{
        transform::visit::{Folder, Visit, Visitor},
        AnnotatedModuleIO, Block, ExprInInstantiatedModule, ExternModuleBody,
        ExternModuleParameter, ExternModuleParameterValue, InstantiatedModule, ModuleBody,
        ModuleIO, NameId, NormalModuleBody, Stmt, StmtConnect, StmtDeclaration, StmtFormal, StmtIf,
        StmtInstance, StmtMatch, StmtReg, StmtWire,
    },
    prelude::*,
    source_location,
};
use std::{
    borrow::Borrow,
    convert::Infallible,
    fmt,
    hash::{Hash, Hasher},
    marker::PhantomData,
    mem,
};

#[derive(Debug)]
pub enum DeduceResetsError {
    ResetIsNotDrivenByAsyncOrSync { source_location: SourceLocation },
    ResetIsDrivenByBothAsyncAndSync { source_location: SourceLocation },
}

impl fmt::Display for DeduceResetsError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            DeduceResetsError::ResetIsNotDrivenByAsyncOrSync { source_location } => {
                write!(f, "deduce_reset failed: Reset signal is not driven by any AsyncReset or SyncReset signals: {source_location}")
            }
            DeduceResetsError::ResetIsDrivenByBothAsyncAndSync { source_location } => {
                write!(f, "deduce_reset failed: Reset signal is driven by both AsyncReset and SyncReset signals: {source_location}")
            }
        }
    }
}

impl std::error::Error for DeduceResetsError {}

impl From<DeduceResetsError> for std::io::Error {
    fn from(value: DeduceResetsError) -> Self {
        std::io::Error::new(std::io::ErrorKind::Other, value)
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
enum ResetsLayout {
    NoResets,
    Reset,
    SyncReset,
    AsyncReset,
    Bundle {
        fields: Interned<[ResetsLayout]>,
        reset_count: usize,
    },
    Enum {
        variants: Interned<[ResetsLayout]>,
        reset_count: usize,
    },
    Array {
        element: Interned<ResetsLayout>,
        reset_count: usize,
    },
}

impl ResetsLayout {
    fn reset_count(self) -> usize {
        match self {
            ResetsLayout::NoResets => 0,
            ResetsLayout::Reset | ResetsLayout::SyncReset | ResetsLayout::AsyncReset => 1,
            ResetsLayout::Bundle { reset_count, .. }
            | ResetsLayout::Enum { reset_count, .. }
            | ResetsLayout::Array { reset_count, .. } => reset_count,
        }
    }
    fn new(ty: CanonicalType) -> Self {
        #[derive(Clone, Copy, PartialEq, Eq, Hash)]
        struct MyMemoize;
        impl Memoize for MyMemoize {
            type Input = CanonicalType;
            type InputOwned = CanonicalType;
            type Output = ResetsLayout;

            fn inner(self, ty: &Self::Input) -> Self::Output {
                match *ty {
                    CanonicalType::UInt(_) => ResetsLayout::NoResets,
                    CanonicalType::SInt(_) => ResetsLayout::NoResets,
                    CanonicalType::Bool(_) => ResetsLayout::NoResets,
                    CanonicalType::Array(ty) => {
                        let element = ResetsLayout::new(ty.element()).intern_sized();
                        ResetsLayout::Array {
                            element,
                            reset_count: element.reset_count(),
                        }
                    }
                    CanonicalType::Enum(ty) => {
                        let mut reset_count = 0;
                        let variants = Interned::from_iter(ty.variants().iter().map(|variant| {
                            let resets_layout =
                                variant.ty.map_or(ResetsLayout::NoResets, ResetsLayout::new);
                            reset_count += resets_layout.reset_count();
                            resets_layout
                        }));
                        ResetsLayout::Enum {
                            variants,
                            reset_count,
                        }
                    }
                    CanonicalType::Bundle(ty) => {
                        let mut reset_count = 0;
                        let fields = Interned::from_iter(ty.fields().iter().map(|field| {
                            let resets_layout = ResetsLayout::new(field.ty);
                            reset_count += resets_layout.reset_count();
                            resets_layout
                        }));
                        ResetsLayout::Bundle {
                            fields,
                            reset_count,
                        }
                    }
                    CanonicalType::AsyncReset(_) => ResetsLayout::AsyncReset,
                    CanonicalType::SyncReset(_) => ResetsLayout::SyncReset,
                    CanonicalType::Reset(_) => ResetsLayout::Reset,
                    CanonicalType::Clock(_) => ResetsLayout::NoResets,
                }
            }
        }
        MyMemoize.get_owned(ty)
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
struct ResetNodeIndex(usize);

#[derive(Copy, Clone, Debug)]
struct ResetNode {
    is_async: Option<bool>,
    source_location: Option<SourceLocation>,
}

impl ResetNode {
    fn union(
        self,
        other: Self,
        fallback_error_source_location: SourceLocation,
    ) -> Result<Self, DeduceResetsError> {
        match (self.is_async, other.is_async) {
            (None, None) => Ok(Self {
                is_async: None,
                source_location: self.source_location.or(other.source_location),
            }),
            (None, is_async @ Some(_)) => Ok(Self {
                is_async,
                // prioritize `other`
                source_location: other.source_location.or(self.source_location),
            }),
            (is_async @ Some(_), None) => Ok(Self {
                is_async,
                // prioritize `self`
                source_location: self.source_location.or(other.source_location),
            }),
            (Some(self_is_async), Some(other_is_async)) => {
                if self_is_async == other_is_async {
                    Ok(Self {
                        is_async: Some(self_is_async),
                        source_location: self.source_location.or(other.source_location),
                    })
                } else {
                    Err(DeduceResetsError::ResetIsDrivenByBothAsyncAndSync {
                        source_location: self
                            .source_location
                            .or(other.source_location)
                            .unwrap_or(fallback_error_source_location),
                    })
                }
            }
        }
    }
}

#[derive(Debug, Default)]
struct ResetGraph {
    union_find: UnionFind<usize>,
    nodes: Vec<ResetNode>,
}

impl ResetGraph {
    fn new_node(
        &mut self,
        is_async: Option<bool>,
        source_location: Option<SourceLocation>,
    ) -> ResetNodeIndex {
        let index = self.union_find.new_set();
        assert_eq!(index, self.nodes.len());
        self.nodes.push(ResetNode {
            is_async,
            source_location,
        });
        ResetNodeIndex(index)
    }
    fn union(
        &mut self,
        a: ResetNodeIndex,
        b: ResetNodeIndex,
        fallback_error_source_location: SourceLocation,
    ) -> Result<(), DeduceResetsError> {
        let a = self.union_find.find_mut(a.0);
        let b = self.union_find.find_mut(b.0);
        if a != b {
            self.union_find.union(a, b);
            let merged = self.union_find.find_mut(a);
            self.nodes[merged] =
                self.nodes[a].union(self.nodes[b], fallback_error_source_location)?;
        }
        Ok(())
    }
    fn is_async(
        &mut self,
        node: ResetNodeIndex,
        fallback_to_sync_reset: bool,
        fallback_error_source_location: SourceLocation,
    ) -> Result<bool, DeduceResetsError> {
        let ResetNode {
            is_async,
            source_location,
        } = self.nodes[self.union_find.find_mut(node.0)];
        if let Some(is_async) = is_async {
            Ok(is_async)
        } else if fallback_to_sync_reset {
            Ok(false)
        } else {
            Err(DeduceResetsError::ResetIsNotDrivenByAsyncOrSync {
                source_location: source_location.unwrap_or(fallback_error_source_location),
            })
        }
    }
}

#[derive(Debug)]
enum MaybeInterned<'a, T: ?Sized + Intern> {
    Interned(Interned<T>),
    Borrowed(&'a T),
}

impl<'a, T: ?Sized + Intern> Borrow<T> for MaybeInterned<'a, T> {
    fn borrow(&self) -> &T {
        &**self
    }
}

impl<'a, T: ?Sized + Intern + PartialEq> PartialEq for MaybeInterned<'a, T> {
    fn eq(&self, other: &Self) -> bool {
        **self == **other
    }
}

impl<'a, T: ?Sized + Intern + Eq> Eq for MaybeInterned<'a, T> {}

impl<'a, T: ?Sized + Intern + Hash> Hash for MaybeInterned<'a, T> {
    fn hash<H: Hasher>(&self, state: &mut H) {
        (**self).hash(state);
    }
}

impl<'a, T: ?Sized + Intern> Copy for MaybeInterned<'a, T> {}

impl<'a, T: ?Sized + Intern> Clone for MaybeInterned<'a, T> {
    fn clone(&self) -> Self {
        *self
    }
}

impl<'a, T: ?Sized + Intern> std::ops::Deref for MaybeInterned<'a, T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        match self {
            MaybeInterned::Interned(v) => v,
            MaybeInterned::Borrowed(v) => v,
        }
    }
}

impl<'a, T: ?Sized + Intern> MaybeInterned<'a, T> {
    fn to_interned(self) -> Interned<T> {
        match self {
            MaybeInterned::Interned(v) => v,
            MaybeInterned::Borrowed(v) => v.intern(),
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
struct Resets<'a> {
    ty: CanonicalType,
    layout: ResetsLayout,
    node_indexes: MaybeInterned<'a, [ResetNodeIndex]>,
}

impl<'a> Resets<'a> {
    fn into_interned(self) -> Resets<'static> {
        let Self {
            ty,
            layout,
            node_indexes,
        } = self;
        Resets {
            ty,
            layout,
            node_indexes: MaybeInterned::Interned(node_indexes.to_interned()),
        }
    }
}

#[derive(Debug)]
struct State {
    base_module: Interned<Module<Bundle>>,
    expr_resets: HashMap<ExprInInstantiatedModule<CanonicalType>, Resets<'static>>,
    reset_graph: ResetGraph,
    fallback_to_sync_reset: bool,
}

struct PassOutput<T, P: Pass>(P::Output<T>);

impl<T, P: Pass> PassOutput<T, P> {
    fn new(v: T) -> Self {
        P::output_new(v)
    }
    fn from_fn(f: impl FnOnce() -> T) -> Self {
        PassOutput::new(()).map(|()| f())
    }
    fn as_ref(&self) -> PassOutput<&T, P> {
        P::output_as_ref(self)
    }
    fn as_mut(&mut self) -> PassOutput<&mut T, P> {
        P::output_as_mut(self)
    }
    fn map<U>(self, f: impl FnOnce(T) -> U) -> PassOutput<U, P> {
        P::map(self, f)
    }
}

trait PassOutputZip<P: Pass>: Sized {
    type Zipped;
    fn zip(self) -> PassOutput<Self::Zipped, P>;
    fn call<U>(self, f: impl FnOnce(Self::Zipped) -> U) -> PassOutput<U, P> {
        self.zip().map(f)
    }
}

impl<P: Pass> PassOutputZip<P> for () {
    type Zipped = ();
    fn zip(self) -> PassOutput<Self::Zipped, P> {
        PassOutput::new(())
    }
}

impl<T, P: Pass> PassOutputZip<P> for (PassOutput<T, P>,) {
    type Zipped = (T,);
    fn zip(self) -> PassOutput<Self::Zipped, P> {
        self.0.map(|v| (v,))
    }
}

macro_rules! impl_zip {
    ($first_arg:ident: $first_T:ident, $($arg:ident: $T:ident),* $(,)?) => {
        impl_zip!(@step [], [($first_arg: $first_T) $(($arg: $T))*], (),);
    };
    (
        @impl($first_arg:tt,),
        $tuple_pat:tt,
    ) => {};
    (
        @impl(($first_arg:ident: $first_T:ident),
            $(($arg:ident: $T:ident),)*),
        $tuple_pat:tt,
    ) => {
        impl<$first_T, $($T,)* P: Pass> PassOutputZip<P> for (PassOutput<$first_T, P>, $(PassOutput<$T, P>),*) {
            type Zipped = ($first_T, $($T),*);
            fn zip(self) -> PassOutput<($first_T, $($T),*), P> {
                let (tuples, $($arg),*) = self;
                $(let tuples = P::zip(tuples, $arg);)*
                tuples.map(|$tuple_pat| ($first_arg, $($arg),*))
            }
        }
    };
    (
        @step [$($cur:tt)*],
        [],
        $tuple_pat:tt,
    ) => {};
    (
        @step [$($cur:tt)*],
        [($next_arg:ident: $next_T:ident) $($rest:tt)*],
        (),
    ) => {
        impl_zip!(@impl($($cur,)* ($next_arg: $next_T),), $next_arg,);
        impl_zip!(@step [$($cur)* ($next_arg: $next_T)], [$($rest)*], $next_arg,);
    };
    (
        @step [$($cur:tt)*],
        [($next_arg:ident: $next_T:ident) $($rest:tt)*],
        $tuple_pat:tt,
    ) => {
        impl_zip!(@impl($($cur,)* ($next_arg: $next_T),), ($tuple_pat, $next_arg),);
        impl_zip!(@step [$($cur)* ($next_arg: $next_T)], [$($rest)*], ($tuple_pat, $next_arg),);
    };
}

impl_zip!(t0: T0, t1: T1, t2: T2, t3: T3, t4: T4, t5: T5, t6: T6, t7: T7, t8: T8, t9: T9, t10: T10, t11: T11);

impl<T: FromIterator<A>, P: Pass, A> FromIterator<PassOutput<A, P>> for PassOutput<T, P> {
    fn from_iter<I: IntoIterator<Item = PassOutput<A, P>>>(iter: I) -> Self {
        P::output_from_iter(iter)
    }
}

trait Pass: Sized {
    type Output<T>;
    fn output_new<T>(v: T) -> PassOutput<T, Self>;
    fn output_as_ref<T>(v: &PassOutput<T, Self>) -> PassOutput<&T, Self>;
    fn output_as_mut<T>(v: &mut PassOutput<T, Self>) -> PassOutput<&mut T, Self>;
    fn output_from_iter<T: FromIterator<A>, A>(
        iter: impl IntoIterator<Item = PassOutput<A, Self>>,
    ) -> PassOutput<T, Self>;
    fn map<T, U>(v: PassOutput<T, Self>, f: impl FnOnce(T) -> U) -> PassOutput<U, Self>;
    fn zip<T, U>(t: PassOutput<T, Self>, u: PassOutput<U, Self>) -> PassOutput<(T, U), Self>;
}

struct BuildResetGraph;

impl Pass for BuildResetGraph {
    type Output<T> = ();

    fn output_new<T>(_v: T) -> PassOutput<T, Self> {
        PassOutput(())
    }

    fn output_as_ref<T>(_v: &PassOutput<T, Self>) -> PassOutput<&T, Self> {
        PassOutput(())
    }

    fn output_as_mut<T>(_v: &mut PassOutput<T, Self>) -> PassOutput<&mut T, Self> {
        PassOutput(())
    }

    fn output_from_iter<T: FromIterator<A>, A>(
        iter: impl IntoIterator<Item = PassOutput<A, Self>>,
    ) -> PassOutput<T, Self> {
        iter.into_iter().for_each(|_| {});
        PassOutput(())
    }

    fn map<T, U>(_v: PassOutput<T, Self>, _f: impl FnOnce(T) -> U) -> PassOutput<U, Self> {
        PassOutput(())
    }

    fn zip<T, U>(_t: PassOutput<T, Self>, _u: PassOutput<U, Self>) -> PassOutput<(T, U), Self> {
        PassOutput(())
    }
}

struct SubstituteResets;

impl Pass for SubstituteResets {
    type Output<T> = T;

    fn output_new<T>(v: T) -> PassOutput<T, Self> {
        PassOutput(v)
    }

    fn output_as_ref<T>(v: &PassOutput<T, Self>) -> PassOutput<&T, Self> {
        PassOutput(&v.0)
    }

    fn output_as_mut<T>(v: &mut PassOutput<T, Self>) -> PassOutput<&mut T, Self> {
        PassOutput(&mut v.0)
    }

    fn output_from_iter<T: FromIterator<A>, A>(
        iter: impl IntoIterator<Item = PassOutput<A, Self>>,
    ) -> PassOutput<T, Self> {
        PassOutput(T::from_iter(iter.into_iter().map(|PassOutput(v)| v)))
    }

    fn map<T, U>(v: PassOutput<T, Self>, f: impl FnOnce(T) -> U) -> PassOutput<U, Self> {
        PassOutput(f(v.0))
    }

    fn zip<T, U>(t: PassOutput<T, Self>, u: PassOutput<U, Self>) -> PassOutput<(T, U), Self> {
        PassOutput((t.0, u.0))
    }
}

struct PassArgs<'a, P: Pass> {
    state: &'a mut State,
    instantiated_module: InstantiatedModule,
    fallback_error_source_location: SourceLocation,
    _phantom: PhantomData<P>,
}

impl<'a, P: Pass> fmt::Debug for PassArgs<'a, P> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            state,
            instantiated_module,
            fallback_error_source_location,
            _phantom: _,
        } = self;
        f.debug_struct("PassArgs")
            .field("state", state)
            .field("instantiated_module", instantiated_module)
            .field(
                "fallback_error_source_location",
                fallback_error_source_location,
            )
            .finish()
    }
}

impl<'a, P: Pass> PassArgs<'a, P> {
    fn as_mut(&mut self) -> PassArgs<'_, P> {
        let PassArgs {
            ref mut state,
            instantiated_module,
            fallback_error_source_location,
            _phantom: _,
        } = *self;
        PassArgs {
            state: &mut **state,
            instantiated_module,
            fallback_error_source_location,
            _phantom: PhantomData,
        }
    }
}

trait RunPass<P: Pass>: Sized {
    fn run_pass(
        &self,
        pass_args: PassArgs<'_, P>,
    ) -> Result<PassOutput<Self, P>, DeduceResetsError>;
}

impl<T: RunPass<P> + Intern + Clone, P: Pass> RunPass<P> for Interned<T> {
    fn run_pass(
        &self,
        pass_args: PassArgs<'_, P>,
    ) -> Result<PassOutput<Self, P>, DeduceResetsError> {
        Ok(T::run_pass(self, pass_args)?.map(Intern::intern_sized))
    }
}

impl<T: RunPass<P> + Clone, P: Pass> RunPass<P> for Interned<[T]>
where
    [T]: Intern,
{
    fn run_pass(
        &self,
        mut pass_args: PassArgs<'_, P>,
    ) -> Result<PassOutput<Self, P>, DeduceResetsError> {
        Result::from_iter(self.iter().map(|v| v.run_pass(pass_args.as_mut())))
    }
}

macro_rules! impl_run_pass_copy {
    ([$($generics:tt)*] $ty:ty) => {
        impl<P: Pass, $($generics)*> RunPass<P> for $ty {
            fn run_pass(
                &self,
                _pass_args: PassArgs<'_, P>,
            ) -> Result<PassOutput<Self, P>, DeduceResetsError> {
                Ok(PassOutput::new(*self))
            }
        }
    };
}

macro_rules! impl_run_pass_clone {
    ([$($generics:tt)*] $ty:ty) => {
        impl<P: Pass, $($generics)*> RunPass<P> for $ty {
            fn run_pass(
                &self,
                _pass_args: PassArgs<'_, P>,
            ) -> Result<PassOutput<Self, P>, DeduceResetsError> {
                Ok(PassOutput::from_fn(|| self.clone()))
            }
        }
    };
}

impl_run_pass_copy!([] Interned<str>);
impl_run_pass_copy!([] usize);
impl_run_pass_copy!([] bool);
impl_run_pass_clone!([] BigInt);

macro_rules! impl_run_pass_for_struct {
    (
        $(#[adjust_pass_args = $adjust_pass_args:expr])?
        #[constructor = $constructor:expr]
        impl[$($generics:tt)*] $RunPass:ident for $ty:ty {
            $($field:ident $(($($args:tt)*))?: _,)*
        }
    ) => {
        impl<P: Pass, $($generics)*> $RunPass<P> for $ty {
            #[allow(unused_mut, unused_variables)]
            fn run_pass(
                &self,
                mut pass_args: PassArgs<'_, P>,
            ) -> Result<PassOutput<Self, P>, DeduceResetsError> {
                $($adjust_pass_args(self, &mut pass_args);)?
                Ok(($(self.$field$(($($args)*))?.run_pass(pass_args.as_mut())?,)*).call(|($($field,)*)| $constructor))
            }
        }
    };
    (
        $(#[adjust_pass_args = $adjust_pass_args:expr])?
        impl[$($generics:tt)*] $RunPass:ident for $ty:ty {
            $($field:ident: _,)*
        }
    ) => {
        impl_run_pass_for_struct! {
            $(#[adjust_pass_args = $adjust_pass_args])?
            #[constructor = {
                type Ty<T> = T;
                Ty::<$ty> { $($field,)* }
            }]
            impl[$($generics)*] $RunPass for $ty {
                $($field: _,)*
            }
        }
    };
}

macro_rules! impl_run_pass_for_enum {
    (impl[$($generics:tt)*] $RunPass:ident for $ty:ty {
        $($variant:ident($arg:ident),)*
    }) => {
        impl<P: Pass, $($generics)*> $RunPass<P> for $ty {
            fn run_pass(
                &self,
                pass_args: PassArgs<'_, P>,
            ) -> Result<PassOutput<Self, P>, DeduceResetsError> {
                type Ty<T> = T;
                match self {
                    $(Ty::<$ty>::$variant($arg) => Ok($arg.run_pass(pass_args)?.map(<$ty>::$variant)),)*
                }
            }
        }
    };
}

impl_run_pass_for_struct! {
    impl[] RunPass for Block {
        memories: _,
        stmts: _,
    }
}

impl_run_pass_for_struct! {
    impl[] RunPass for NormalModuleBody {
        body: _,
    }
}

impl_run_pass_copy!([] ExternModuleBody);
impl_run_pass_copy!([] NameId);
impl_run_pass_copy!([] SourceLocation);
impl_run_pass_copy!([] DontTouchAnnotation);
impl_run_pass_copy!([] SVAttributeAnnotation);
impl_run_pass_copy!([] BlackBoxInlineAnnotation);
impl_run_pass_copy!([] BlackBoxPathAnnotation);
impl_run_pass_copy!([] DocStringAnnotation);
impl_run_pass_copy!([] CustomFirrtlAnnotation);
impl_run_pass_copy!([] MemPort<DynPortType>); // Mem can't contain any `Reset` types
impl_run_pass_copy!([] Mem); // Mem can't contain any `Reset` types

impl_run_pass_for_enum! {
    impl[] RunPass for TargetBase {
        ModuleIO(v),
        MemPort(v),
        Reg(v),
        Wire(v),
        Instance(v),
    }
}

impl_run_pass_for_struct! {
    impl[] RunPass for TargetPathBundleField {
        name: _,
    }
}

impl_run_pass_for_struct! {
    impl[] RunPass for TargetPathArrayElement {
        index: _,
    }
}

impl_run_pass_for_struct! {
    impl[] RunPass for TargetPathDynArrayElement {}
}

impl_run_pass_for_enum! {
    impl[] RunPass for TargetPathElement {
        BundleField(v),
        ArrayElement(v),
        DynArrayElement(v),
    }
}

impl_run_pass_for_enum! {
    impl[] RunPass for Target {
        Base(v),
        Child(v),
    }
}

impl_run_pass_for_struct! {
    #[constructor = TargetChild::new(parent, path_element)]
    impl[] RunPass for TargetChild {
        parent(): _,
        path_element(): _,
    }
}

impl_run_pass_for_enum! {
    impl[] RunPass for Annotation {
        DontTouch(v),
        SVAttribute(v),
        BlackBoxInline(v),
        BlackBoxPath(v),
        DocString(v),
        CustomFirrtl(v),
    }
}

impl_run_pass_for_enum! {
    impl[] RunPass for ModuleBody {
        Normal(v),
        Extern(v),
    }
}

impl_run_pass_for_struct! {
    #[constructor = TargetedAnnotation::new(target, annotation)]
    impl[] RunPass for TargetedAnnotation {
        target(): _,
        annotation(): _,
    }
}

impl_run_pass_for_struct! {
    impl[] RunPass for AnnotatedModuleIO {
        annotations: _,
        module_io: _,
    }
}

impl_run_pass_for_struct! {
    #[adjust_pass_args = |module: &Module<_>, pass_args: &mut PassArgs<'_, _>| {
        pass_args.fallback_error_source_location = module.source_location();
    }]
    #[constructor = Module::new_unchecked(
        name_id,
        source_location,
        body,
        module_io,
        module_annotations,
    )]
    impl[] RunPass for Module<Bundle> {
        name_id(): _,
        source_location(): _,
        body(): _,
        module_io(): _,
        module_annotations(): _,
    }
}

pub fn deduce_resets(
    module: Interned<Module<Bundle>>,
    fallback_to_sync_reset: bool,
) -> Result<Interned<Module<Bundle>>, DeduceResetsError> {
    let mut state = State {
        base_module: module,
        expr_resets: HashMap::new(),
        reset_graph: ResetGraph::default(),
        fallback_to_sync_reset,
    };
    RunPass::<BuildResetGraph>::run_pass(
        &*module,
        PassArgs {
            state: &mut state,
            instantiated_module: InstantiatedModule::Base(module),
            fallback_error_source_location: module.source_location(),
            _phantom: PhantomData,
        },
    )?;
    Ok(RunPass::<SubstituteResets>::run_pass(
        &*module,
        PassArgs {
            state: &mut state,
            instantiated_module: InstantiatedModule::Base(module),
            fallback_error_source_location: module.source_location(),
            _phantom: PhantomData,
        },
    )?
    .0
    .intern_sized())
}
